from abc import abstractmethod
from typing import Generic, TypeVar
from dataclasses import field, dataclass

from .abc import Attribute
from .types import TypeAttribute

T = TypeVar("T", covariant=True)


@dataclass(eq=False)
class Data(Attribute, Generic[T]):
    """Base class for data attributes.

    Data attributes are compile-time constants that can be used to
    represent runtime data inside the IR.

    This class is meant to be subclassed by specific data attributes.
    It provides a `type` attribute that should be set to the type of
    the data.
    """

    type: TypeAttribute = field(init=False, repr=False)

    @abstractmethod
    def unwrap(self) -> T:
        """Returns the underlying data value."""
        ...



import ast

from kirin import types, lowering

from .stmts import New
from ._dialect import dialect


@dialect.register
class PythonLowering(lowering.FromPythonAST):

    def lower_List(self, state: lowering.State, node: ast.List) -> lowering.Result:
        elts = tuple(state.lower(each).expect_one() for each in node.elts)

        if len(elts):
            typ = elts[0].type
            for each in elts:
                typ = typ.join(each.type)
        else:
            typ = types.Any

        return state.current_frame.push(New(values=tuple(elts)))



from kirin.lattice.abc import SingletonMeta, BoundedLattice


class EmptyLattice(BoundedLattice["EmptyLattice"], metaclass=SingletonMeta):
    """Empty lattice."""

    def join(self, other: "EmptyLattice") -> "EmptyLattice":
        return self

    def meet(self, other: "EmptyLattice") -> "EmptyLattice":
        return self

    @classmethod
    def bottom(cls):
        return cls()

    @classmethod
    def top(cls):
        return cls()

    def __hash__(self) -> int:
        return id(self)

    def is_subseteq(self, other: "EmptyLattice") -> bool:
        return True



from typing import TypeVar, final

from kirin import ir, types, interp
from kirin.decl import fields
from kirin.analysis import const
from kirin.interp.impl import Signature
from kirin.analysis.forward import Forward, ForwardFrame

from .solve import TypeResolution


@final
class TypeInference(Forward[types.TypeAttribute]):
    """Type inference analysis for kirin.

    This analysis uses the forward dataflow analysis framework to infer the types of
    the IR. The analysis uses the type information within the IR to determine the
    method dispatch.

    The analysis will fallback to a type resolution algorithm if the type information
    is not available in the IR but the type information is available in the abstract
    values.
    """

    keys = ["typeinfer"]
    lattice = types.TypeAttribute

    def run_analysis(
        self,
        method: ir.Method,
        args: tuple[types.TypeAttribute, ...] | None = None,
        *,
        no_raise: bool = True,
    ) -> tuple[ForwardFrame[types.TypeAttribute], types.TypeAttribute]:
        if args is None:
            args = method.arg_types
        return super().run_analysis(method, args, no_raise=no_raise)

    # NOTE: unlike concrete interpreter, instead of using type information
    # within the IR. Type inference will use the interpreted
    # value (which is a type) to determine the method dispatch.
    def build_signature(
        self, frame: ForwardFrame[types.TypeAttribute], stmt: ir.Statement
    ) -> Signature:
        _args = ()
        for x in frame.get_values(stmt.args):
            # TODO: remove this after we have multiple dispatch...
            if isinstance(x, types.Generic):
                _args += (x.body,)
            else:
                _args += (x,)
        return Signature(stmt.__class__, _args)

    def eval_stmt_fallback(
        self, frame: ForwardFrame[types.TypeAttribute], stmt: ir.Statement
    ) -> tuple[types.TypeAttribute, ...] | interp.SpecialValue[types.TypeAttribute]:
        resolve = TypeResolution()
        fs = fields(stmt)
        for f, value in zip(fs.args.values(), frame.get_values(stmt.args)):
            resolve.solve(f.type, value)
        for arg, f in zip(stmt.args, fs.args.values()):
            frame.set(arg, frame.get(arg).meet(resolve.substitute(f.type)))
        return tuple(resolve.substitute(result.type) for result in stmt.results)

    def run_method(
        self, method: ir.Method, args: tuple[types.TypeAttribute, ...]
    ) -> tuple[ForwardFrame[types.TypeAttribute], types.TypeAttribute]:
        return self.run_callable(method.code, (method.self_type,) + args)

    T = TypeVar("T")

    @classmethod
    def maybe_const(cls, value: ir.SSAValue, type_: type[T]) -> T | None:
        """Get a constant value of a given type.

        If the value is not a constant or the constant is not of the given type, return
        `None`.
        """
        hint = value.hints.get("const")
        if isinstance(hint, const.Value) and isinstance(hint.data, type_):
            return hint.data

    @classmethod
    def expect_const(cls, value: ir.SSAValue, type_: type[T]):
        """Expect a constant value of a given type.

        If the value is not a constant or the constant is not of the given type, raise
        an `InterpreterError`.
        """
        hint = cls.maybe_const(value, type_)
        if hint is None:
            raise interp.InterpreterError(f"expected {type_}, got {hint}")
        return hint



from dataclasses import field, dataclass

from kirin.ir import Method, HasSignature
from kirin.rewrite import Walk, Chain
from kirin.passes.abc import Pass
from kirin.rewrite.abc import RewriteResult
from kirin.dialects.func import Signature
from kirin.analysis.typeinfer import TypeInference
from kirin.rewrite.apply_type import ApplyType
from kirin.rewrite.type_assert import InlineTypeAssert

from .hint_const import HintConst
from .post_inference import PostInference


@dataclass
class TypeInfer(Pass):
    hint_const: HintConst = field(init=False)
    inference: PostInference = field(init=False)

    def __post_init__(self):
        self.infer = TypeInference(self.dialects)
        self.hint_const = HintConst(self.dialects, no_raise=self.no_raise)
        self.post_inference = PostInference(self.dialects, no_raise=self.no_raise)

    def unsafe_run(self, mt: Method) -> RewriteResult:
        result = self.hint_const.unsafe_run(mt)
        frame, return_type = self.infer.run_analysis(
            mt, mt.arg_types, no_raise=self.no_raise
        )
        if trait := mt.code.get_trait(HasSignature):
            trait.set_signature(mt.code, Signature(mt.arg_types, return_type))

        result = (
            Chain(
                Walk(ApplyType(frame.entries)),
                Walk(InlineTypeAssert()),
            )
            .rewrite(mt.code)
            .join(result)
        )
        result = self.post_inference.fixpoint(mt).join(result)
        mt.inferred = True
        return result



import sys
from abc import ABC
from typing import TypeVar, Iterable
from dataclasses import dataclass

from kirin import ir, interp
from kirin.interp import AbstractFrame, AbstractInterpreter
from kirin.lattice import BoundedLattice

ExtraType = TypeVar("ExtraType")
LatticeElemType = TypeVar("LatticeElemType", bound=BoundedLattice)


@dataclass
class ForwardFrame(AbstractFrame[LatticeElemType]):
    pass


ForwardFrameType = TypeVar("ForwardFrameType", bound=ForwardFrame)


@dataclass
class ForwardExtra(
    AbstractInterpreter[ForwardFrameType, LatticeElemType],
    ABC,
):
    """Abstract interpreter but record results for each SSA value.

    Params:
        LatticeElemType: The lattice element type.
        ExtraType: The type of extra information to be stored in the frame.
    """

    def run_analysis(
        self,
        method: ir.Method,
        args: tuple[LatticeElemType, ...] | None = None,
        *,
        no_raise: bool = True,
    ) -> tuple[ForwardFrameType, LatticeElemType]:
        """Run the forward dataflow analysis.

        Args:
            method(ir.Method): The method to analyze.
            args(tuple[LatticeElemType]): The arguments to the method. Defaults to tuple of top values.

        Keyword Args:
            no_raise(bool): If True, return bottom values if the analysis fails. Defaults to True.

        Returns:
            ForwardFrameType: The results of the analysis contained in the frame.
            LatticeElemType: The result of the analysis for the method return value.
        """
        args = args or tuple(self.lattice.top() for _ in method.args)

        if self._eval_lock:
            raise interp.InterpreterError(
                "recursive eval is not allowed, use run_method instead"
            )

        self._eval_lock = True
        self.initialize()
        current_recursion_limit = sys.getrecursionlimit()
        sys.setrecursionlimit(self.max_python_recursion_depth)
        try:
            frame, ret = self.run_method(method, args)
        except Exception as e:
            # NOTE: initialize will create new State
            # so we don't need to copy the frames.
            if not no_raise:
                raise e
            return self.state.current_frame, self.lattice.bottom()
        finally:
            self._eval_lock = False
            sys.setrecursionlimit(current_recursion_limit)
        return frame, ret

    def set_values(
        self,
        frame: AbstractFrame[LatticeElemType],
        ssa: Iterable[ir.SSAValue],
        results: Iterable[LatticeElemType],
    ):
        """Set the abstract values for the given SSA values in the frame.

        This method is used to customize how the abstract values are set in
        the frame. By default, the abstract values are set directly in the
        frame. This method is overridden to join the results if the SSA value
        already exists in the frame.
        """
        for ssa_value, result in zip(ssa, results):
            if ssa_value in frame.entries:
                frame.entries[ssa_value] = frame.entries[ssa_value].join(result)
            else:
                frame.entries[ssa_value] = result


class Forward(ForwardExtra[ForwardFrame[LatticeElemType], LatticeElemType], ABC):
    """Forward dataflow analysis.

    This is the base class for forward dataflow analysis. If your analysis
    requires extra information per frame, you should subclass
    [`ForwardExtra`][kirin.analysis.forward.ForwardExtra] instead.
    """

    def initialize_frame(
        self, code: ir.Statement, *, has_parent_access: bool = False
    ) -> ForwardFrame[LatticeElemType]:
        return ForwardFrame(code, has_parent_access=has_parent_access)



"""The range dialect for Python.

This dialect models the builtin `range()` function in Python.

The dialect includes:
- The `Range` statement class.
- The lowering pass for the `range()` function.

This dialect does not include a concrete implementation or type inference
for the `range()` function. One needs to use other dialect for the concrete
implementation and type inference, e.g., `ilist` dialect.
"""

from kirin import ir, types, interp, lowering
from kirin.decl import info, statement
from kirin.dialects import eltype

dialect = ir.Dialect("py.range")


@statement(dialect=dialect)
class Range(ir.Statement):
    name = "range"
    traits = frozenset({ir.Pure(), lowering.FromPythonRangeLike()})
    start: ir.SSAValue = info.argument(types.Int)
    stop: ir.SSAValue = info.argument(types.Int)
    step: ir.SSAValue = info.argument(types.Int)
    result: ir.ResultValue = info.result(types.PyClass(range))


@dialect.register(key="typeinfer")
class TypeInfer(interp.MethodTable):

    @interp.impl(eltype.ElType, types.PyClass(range))
    def eltype_range(self, interp_, frame: interp.Frame, stmt: eltype.ElType):
        return (types.Int,)



"""builtin dialect for python builtins

This dialect provides implementations for builtin functions like abs and sum.

- Statements: `Abs`, `Sum`.
- The lowering pass for the builtin functions.
- The concrete implementation of the builtin functions.
- The type inference implementation of the builtin functions.

This dialect maps `ast.Call` nodes of builtin functions to the `Abs` and `Sum` statements.
"""

from ast import Call

from kirin import ir, types, interp, lowering
from kirin.decl import info, statement

dialect = ir.Dialect("py.builtin")

T = types.TypeVar("T", bound=types.Int | types.Float)


@statement(dialect=dialect)
class Abs(ir.Statement):
    name = "abs"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    value: ir.SSAValue = info.argument(T, print=False)
    result: ir.ResultValue = info.result(T)


@statement(dialect=dialect)
class Sum(ir.Statement):
    name = "sum"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    value: ir.SSAValue = info.argument(types.Any, print=False)
    result: ir.ResultValue = info.result(types.Any)


@dialect.register
class Lowering(lowering.FromPythonAST):

    @lowering.akin(abs)
    def lower_Call_abs(self, state: lowering.State, node: Call) -> lowering.Result:
        return state.current_frame.push(Abs(state.lower(node.args[0]).expect_one()))

    @lowering.akin(sum)
    def lower_Call_sum(self, state: lowering.State, node: Call) -> lowering.Result:
        return state.current_frame.push(Sum(state.lower(node.args[0]).expect_one()))


@dialect.register
class Concrete(interp.MethodTable):

    @interp.impl(Abs)
    def abs(self, interp, frame: interp.Frame, stmt: Abs):
        return (abs(frame.get(stmt.value)),)

    @interp.impl(Sum)
    def _sum(self, interp, frame: interp.Frame, stmt: Sum):
        return (sum(frame.get(stmt.value)),)


@dialect.register(key="typeinfer")
class TypeInfer(interp.MethodTable):

    @interp.impl(Abs, types.Int)
    def absi(self, interp, frame, stmt):
        return (types.Int,)

    @interp.impl(Abs, types.Float)
    def absf(self, interp, frame, stmt):
        return (types.Float,)


dialect.register_py_type(float)
dialect.register_py_type(int)
dialect.register_py_type(bool)
dialect.register_py_type(str)
dialect.register_py_type(type(None))
dialect.register_py_type(list)
dialect.register_py_type(dict)
dialect.register_py_type(tuple)
dialect.register_py_type(set)
dialect.register_py_type(frozenset)
dialect.register_py_type(slice)



from kirin.prelude import basic
from kirin.dialects import ilist


def test_map_wrapper():

    @basic
    def add1(x: int):
        return x + 1

    @basic
    def map_wrap():
        return ilist.map(add1, range(5))

    out = map_wrap()
    assert isinstance(out, ilist.IList)
    assert out.data == [1, 2, 3, 4, 5]


def test_foldr_wrapper():

    @basic
    def add_fold(x: int, out: int):
        return out + x

    @basic
    def map_foldr():
        return ilist.foldr(add_fold, range(5), init=10)

    out = map_foldr()
    assert isinstance(out, int)
    assert out == 10 + 0 + 1 + 2 + 3 + 4


def test_foldl_wrapper():

    @basic
    def add_fold2(out: int, x: int):
        return out + x

    @basic
    def map_foldl():
        return ilist.foldr(add_fold2, range(5), init=10)

    out = map_foldl()
    assert isinstance(out, int)
    assert out == 10 + 0 + 1 + 2 + 3 + 4


def test_scan_wrapper():

    @basic
    def add_scan(out: int, x: int):
        return out + 1, out + x

    @basic
    def scan_wrap():
        return ilist.scan(add_scan, range(5), init=10)

    out = scan_wrap()
    assert isinstance(out, tuple)
    assert len(out) == 2

    res = out[0]
    out_list = out[1]

    assert isinstance(res, int)
    assert res == 10 + 1 * 5

    assert isinstance(out_list, ilist.IList)
    assert out_list.data == [
        10 + 0,
        10 + 1 + 1,
        10 + 1 + 1 + 2,
        10 + 1 + 1 + 1 + 3,
        10 + 1 + 1 + 1 + 1 + 4,
    ]



from kirin import ir, types
from kirin.dialects.py import constant
from kirin.rewrite.abc import RewriteRule, RewriteResult
from kirin.dialects.ilist.stmts import IListType
from kirin.dialects.ilist.runtime import IList

from .._dialect import dialect


@dialect.post_inference
class List2IList(RewriteRule):

    def rewrite_Block(self, node: ir.Block) -> RewriteResult:
        has_done_something = False
        for arg in node.args:
            has_done_something = self._rewrite_SSAValue_type(arg)
        return RewriteResult(has_done_something=has_done_something)

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        has_done_something = False
        if isinstance(node, constant.Constant) and isinstance(node.value, list):
            eltype = self._eltype(node.result.type)
            node.replace_by(
                constant.Constant(value=IList(data=node.value, elem=eltype))
            )

        for result in node.results:
            has_done_something = self._rewrite_SSAValue_type(result)

        return RewriteResult(has_done_something=has_done_something)

    def _rewrite_SSAValue_type(self, value: ir.SSAValue):
        # NOTE: cannot use issubseteq here because type can be Bottom
        if isinstance(value.type, types.Generic) and issubclass(
            value.type.body.typ, list
        ):
            value.type = IListType[value.type.vars[0], types.Any]
            return True

        elif isinstance(value.type, types.PyClass) and issubclass(value.type.typ, list):
            value.type = IListType[types.Any, types.Any]
            return True
        return False

    def _eltype(self, type: types.TypeAttribute):
        if isinstance(type, types.Generic) and issubclass(type.body.typ, (list, IList)):
            return type.vars[0]
        return types.Any



from kirin import ir


def test_hash():
    # Test hash collision for PyAttr with different values
    # This is a regression test for issue #1234
    # where hash(-1) == hash(-2) for PyAttr
    # This is fixed by using the value as bytes
    a = ir.PyAttr(-1)
    b = ir.PyAttr(-2)
    assert hash(a) != hash(b)
    assert hash(a.data) == hash(b.data)

    a = ir.PyAttr(-1.0)
    b = ir.PyAttr(-2.0)
    assert hash(a) != hash(b)
    assert hash(a.data) == hash(b.data)



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING

# NOTE: we don't want to actually load rich here
if TYPE_CHECKING:
    from typing import Any, Literal, TypedDict

    from rich.theme import Theme
    from rich.console import Console
    from typing_extensions import Unpack

    from kirin import ir
    from kirin.print import Printer

    class _PrintOptions(TypedDict, total=False):
        console: Console
        hint: str
        analysis: dict["ir.SSAValue", Any]
        show_indent_mark: bool
        theme: Theme | dict | Literal["dark", "light"]


class Printable:
    """Base class for all objects that can be pretty printed.

    This class provides an interface for pretty printing objects. The
    `print` method is the entry point for the printing process. The
    `print_impl` method is the implementation of the printing process
    and should be implemented by the derived classes.
    """

    @staticmethod
    def __get_printer(
        printer: Printer | None = None,
        **options: Unpack["_PrintOptions"],
    ) -> Printer:
        if printer is None:
            from kirin.print import Printer

            return Printer(**options)
        return printer

    def pager(
        self,
        printer: Printer | None = None,
        **options: Unpack["_PrintOptions"],
    ) -> None:
        """Pretty print the object with a pager.

        Args:
            printer (Printer):
                `Printer` object to use for printing.
                If None, a new `Printer` object will be created.

        Keyword Args:
            console (rich.Console):
                The console to use for printing. If not provided, a new console
                will be created.
            analysis (dict[ir.SSAValue, Printable]):
                Analysis results to use for printing. If `None`, no analysis results
            show_indent_mark (bool):
                Whether to show the indentation mark.
            theme (Theme | dict | str):
                The theme to use for printing, defaults to "dark".

        !!! note
            This function also accepts all other `rich.console.Console` options.
        """
        printer = self.__get_printer(printer, **options)
        with printer.console.pager(styles=True, links=True):
            self.print(printer)

    def print(
        self,
        printer: Printer | None = None,
        end: str = "\n",
        **options: Unpack["_PrintOptions"],
    ) -> None:
        """
        Entry point of the printing process.

        Args:
            printer (Printer):
                `Printer` object to use for printing.
                If None, a new `Printer` object will be created.

        Keyword Args:
            console (rich.Console):
                The console to use for printing. If not provided, a new console
                will be created.
            analysis (dict[ir.SSAValue, Printable]):
                Analysis results to use for printing. If `None`, no analysis results
            show_indent_mark (bool):
                Whether to show the indentation mark.
            theme (Theme | dict | str):
                The theme to use for printing, defaults to "dark".

        !!! note
            This function also accepts all other `rich.console.Console` options.
        """
        printer = self.__get_printer(printer, **options)
        self.print_impl(printer)
        printer.plain_print(end)

    def print_str(
        self,
        printer: Printer | None = None,
        end: str = "\n",
        **options: Unpack["_PrintOptions"],
    ) -> str:
        """Print the object to a string.

        Args:
            printer (Printer):
                `Printer` object to use for printing.
                If None, a new `Printer` object will be created.

        Keyword Args:
            console (rich.Console):
                The console to use for printing. If not provided, a new console
                will be created.
            analysis (dict[ir.SSAValue, Printable]):
                Analysis results to use for printing. If `None`, no analysis results
            show_indent_mark (bool):
                Whether to show the indentation mark.
            theme (Theme | dict | str):
                The theme to use for printing, defaults to "dark".

        !!! note
            This function also accepts all other `rich.console.Console` options.
        """
        printer = self.__get_printer(printer, **options)
        with printer.string_io() as stream:
            self.print(printer, end=end, **options)
            return stream.getvalue()

    @abstractmethod
    def print_impl(self, printer: Printer) -> None:
        raise NotImplementedError(f"print is not implemented for {type(self)}")



"""Compile-time values in Kirin IR.

This module contains the following:

## `abc` module
`abc.AttributeMeta`: The metaclass for all attributes.
`abc.Attribute`: The base class for all attributes.

## `types` module
`types.TypeAttribute`: The base class for all type attributes.
`types.PyClass`: A type attribute representing a Python class.
`types.TypeVar`: A type attribute representing a type variable.
`types.Literal`: A type attribute representing a literal type.
`types.Generic`: A type attribute representing a generic type.
`types.Union`: A type attribute representing a union type.

## `py` module

`py.PyAttr`: An attribute representing a Python value.
"""



from typing import Iterable

from kirin import ir, types
from kirin.interp import Frame, MethodTable, ReturnValue, impl
from kirin.analysis import const
from kirin.analysis.typeinfer import TypeInference, TypeResolution
from kirin.dialects.func.stmts import (
    Call,
    Invoke,
    Lambda,
    Return,
    GetField,
    ConstantNone,
)
from kirin.dialects.func.dialect import dialect


# NOTE: a lot of the type infer rules are same as the builtin dialect
@dialect.register(key="typeinfer")
class TypeInfer(MethodTable):

    @impl(ConstantNone)
    def const_none(self, interp: TypeInference, frame: Frame, stmt: ConstantNone):
        return (types.NoneType,)

    @impl(Return)
    def return_(
        self, interp: TypeInference, frame: Frame[types.TypeAttribute], stmt: Return
    ) -> ReturnValue:
        if (
            isinstance(hint := stmt.value.hints.get("const"), const.Value)
            and hint.data is not None
        ):
            return ReturnValue(types.Literal(hint.data, frame.get(stmt.value)))
        return ReturnValue(frame.get(stmt.value))

    @impl(Call)
    def call(self, interp: TypeInference, frame: Frame, stmt: Call):
        # give up on dynamic method calls
        mt = interp.maybe_const(stmt.callee, ir.Method)
        if mt is None:  # not a constant method
            return self._solve_method_type(interp, frame, stmt)
        return self._invoke_method(
            interp,
            frame,
            mt,
            stmt.args[1:],
            interp.permute_values(
                mt.arg_names, frame.get_values(stmt.inputs), stmt.kwargs
            ),
        )

    def _solve_method_type(self, interp: TypeInference, frame: Frame, stmt: Call):
        mt_inferred = frame.get(stmt.callee)
        if not isinstance(mt_inferred, types.Generic):
            return (types.Bottom,)

        if len(mt_inferred.vars) != 2:
            return (types.Bottom,)
        args = mt_inferred.vars[0]
        result = mt_inferred.vars[1]
        if not args.is_subseteq(types.Tuple):
            return (types.Bottom,)

        resolve = TypeResolution()
        # NOTE: we are not using [...] below to be compatible with 3.10
        resolve.solve(args, types.Tuple.where(frame.get_values(stmt.inputs)))
        return (resolve.substitute(result),)

    @impl(Invoke)
    def invoke(self, interp: TypeInference, frame: Frame, stmt: Invoke):
        return self._invoke_method(
            interp,
            frame,
            stmt.callee,
            stmt.inputs,
            interp.permute_values(
                stmt.callee.arg_names, frame.get_values(stmt.inputs), stmt.kwargs
            ),
        )

    def _invoke_method(
        self,
        interp: TypeInference,
        frame: Frame,
        mt: ir.Method,
        args: Iterable[ir.SSAValue],
        values: tuple,
    ):
        if mt.inferred:  # so we don't end up in infinite loop
            return (mt.return_type,)

        # NOTE: narrowing the argument type based on method signature
        inputs = tuple(
            typ.meet(input_typ) for typ, input_typ in zip(mt.arg_types, values)
        )

        # NOTE: we use lower bound here because function call contains an
        # implicit type check at call site. This will be validated either compile time
        # or runtime.
        # update the results with the narrowed types
        frame.set_values(args, inputs)
        _, ret = interp.run_method(mt, inputs)
        return (ret,)

    @impl(Lambda)
    def lambda_(
        self, interp_: TypeInference, frame: Frame[types.TypeAttribute], stmt: Lambda
    ):
        body_frame, ret = interp_.run_callable(
            stmt,
            (types.MethodType,)
            + tuple(arg.type for arg in stmt.body.blocks[0].args[1:]),
        )
        argtypes = tuple(arg.type for arg in stmt.body.blocks[0].args[1:])
        ret = types.MethodType[[*argtypes], ret]
        frame.entries.update(body_frame.entries)  # pass results back to upper frame
        self_ = stmt.body.blocks[0].args[0]
        frame.set(self_, ret)
        return (ret,)

    @impl(GetField)
    def getfield(self, interp: TypeInference, frame, stmt: GetField):
        return (stmt.result.type,)



import ast

from kirin import lowering

from . import stmts
from ._dialect import dialect


@dialect.register
class Lowering(lowering.FromPythonAST):

    def lower_BinOp(self, state: lowering.State, node: ast.BinOp) -> lowering.Result:
        lhs = state.lower(node.left).expect_one()
        rhs = state.lower(node.right).expect_one()

        if op := getattr(stmts, node.op.__class__.__name__, None):
            stmt = op(lhs=lhs, rhs=rhs)
        else:
            raise lowering.BuildError(f"unsupported binop {node.op}")
        return state.current_frame.push(stmt)



from kirin import types, interp

from . import stmts
from ._dialect import dialect


@dialect.register(key="typeinfer")
class TypeInfer(interp.MethodTable):

    @interp.impl(stmts.UAdd)
    @interp.impl(stmts.USub)
    def uadd(
        self, interp, frame: interp.Frame[types.TypeAttribute], stmt: stmts.UnaryOp
    ):
        return (frame.get(stmt.value),)

    @interp.impl(stmts.Not)
    def not_(self, interp, frame, stmt: stmts.Not):
        return (types.Bool,)

    @interp.impl(stmts.Invert, types.Int)
    def invert(self, interp, frame, stmt):
        return (types.Int,)



from kirin.prelude import basic_no_opt


@basic_no_opt
def not_in(x, y):
    return x not in y


@basic_no_opt
def in_(x, y):
    return x in y


@basic_no_opt
def is_(x, y):
    return x is y


@basic_no_opt
def is_not(x, y):
    return x is not y


def test_is():
    class Foo:
        pass

    a, b = Foo(), Foo()
    assert is_(a, b) == (a is b)
    assert is_(a, a) == (a is a)


def test_is_not():
    class Foo:
        pass

    a, b = Foo(), Foo()
    assert is_not(a, b) == (a is not b)
    assert is_not(a, a) == (a is not a)


def test_in():
    assert in_(1, [1, 2, 3]) == (1 in [1, 2, 3])
    assert in_(4, [1, 2, 3]) == (4 in [1, 2, 3])
    assert in_("a", "abc") == ("a" in "abc")
    assert in_("d", "abc") == ("d" in "abc")


def test_not_in():
    assert not_in(1, [1, 2, 3]) == (1 not in [1, 2, 3])
    assert not_in(4, [1, 2, 3]) == (4 not in [1, 2, 3])
    assert not_in("a", "abc") == ("a" not in "abc")
    assert not_in("d", "abc") == ("d" not in "abc")



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, Iterator
from dataclasses import field, dataclass
from collections.abc import Sequence

from typing_extensions import Self

from kirin.print import Printer
from kirin.ir.ssa import SSAValue, BlockArgument, DeletedSSAValue
from kirin.source import SourceInfo
from kirin.ir.exception import ValidationError
from kirin.ir.nodes.base import IRNode
from kirin.ir.nodes.view import View, MutableSequenceView

if TYPE_CHECKING:
    from kirin.ir.nodes.stmt import Statement
    from kirin.ir.attrs.types import TypeAttribute
    from kirin.ir.nodes.region import Region


@dataclass
class BlockArguments(MutableSequenceView[tuple, "Block", BlockArgument]):
    """A View object that contains a list of BlockArgument.

    Description:
        This is a proxy object that provide safe API to manipulate the arguments of a Block.


    """

    def append_from(self, typ: TypeAttribute, name: str | None = None) -> BlockArgument:
        """Append a new argument to the Block that this View reference to.

        Description:
            This method will create a new [`BlockArgument`][kirin.ir.BlockArgument] and append it to the argument list
            of the reference `Block`.

        Args:
            typ (TypeAttribute): The type of the argument.
            name (str | None, optional): name of the argument. Defaults to `None`.

        Returns:
            BlockArgument: The newly created [`BlockArgument`][kirin.ir.BlockArgument].

        """
        new_arg = BlockArgument(self.node, len(self.node._args), typ)
        if name:
            new_arg.name = name

        self.node._args += (new_arg,)
        return new_arg

    def insert_from(
        self, idx: int, typ: TypeAttribute, name: str | None = None
    ) -> BlockArgument:
        """Insert a new argument to the Block that this View reference to.

        Description:
            This method will create a new `BlockArgument` and insert it to the argument list
            of the reference Block at the specified index

        Args:
            idx (int): Insert location index.
            typ (TypeAttribute): The type of the argument.
            name (str | None, optional): Name of the argument. Defaults to `None`.

        Returns:
            BlockArgument: The newly created BlockArgument.
        """
        if idx < 0 or idx > len(self.node._args):
            raise ValueError("Invalid index")

        new_arg = BlockArgument(self.node, idx, typ)
        if name:
            new_arg.name = name

        for arg in self.node._args[idx:]:
            arg.index += 1
        self.node._args = self.node._args[:idx] + (new_arg,) + self.node._args[idx:]
        return new_arg

    def delete(self, arg: BlockArgument, safe: bool = True) -> None:
        """Delete a BlockArgument from the Block that this View reference to.


        Args:
            arg (BlockArgument): _description_
            safe (bool, optional): If True, error will be raised if the BlockArgument has any Use by others.  Defaults to True.

        Raises:
            ValueError: If the argument does not belong to the reference block.
        """
        if safe and len(arg.uses) > 0:
            raise ValueError("Cannot delete SSA value with uses")

        if arg.block is not self.node:
            raise ValueError("Attempt to delete an argument that is not in the block")

        for block_arg in self.field[arg.index + 1 :]:
            block_arg.index -= 1
        self.node._args = (*self.field[: arg.index], *self.field[arg.index + 1 :])
        arg.replace_by(DeletedSSAValue(arg))

    def __delitem__(self, idx: int) -> None:
        self.delete(self.field[idx])


@dataclass
class BlockStmtIterator:
    """Proxy object to iterate over the Statements in a Block."""

    next_stmt: Statement | None

    def __iter__(self) -> BlockStmtIterator:
        return self

    def __next__(self) -> Statement:
        if self.next_stmt is None:
            raise StopIteration
        stmt = self.next_stmt
        self.next_stmt = stmt.next_stmt
        return stmt


@dataclass
class BlockStmtsReverseIterator:
    """Proxy object to iterate over the Statements in a Block in reverse order."""

    next_stmt: Statement | None

    def __iter__(self) -> BlockStmtsReverseIterator:
        return self

    def __next__(self) -> Statement:
        if self.next_stmt is None:
            raise StopIteration
        stmt = self.next_stmt
        self.next_stmt = stmt.prev_stmt
        return stmt


@dataclass
class BlockStmts(View["Block", "Statement"]):
    """A View object that contains a list of Statements.

    Description:
        This is a proxy object that provide safe API to manipulate the statements of a Block.
    """

    def __iter__(self) -> Iterator[Statement]:
        return BlockStmtIterator(self.node.first_stmt)

    def __len__(self) -> int:
        return self.node._stmt_len

    def __reversed__(self) -> Iterator[Statement]:
        return BlockStmtsReverseIterator(self.node.last_stmt)

    def __repr__(self) -> str:
        return f"BlockStmts(len={len(self)})"

    def __getitem__(self, index: int) -> Statement:
        raise NotImplementedError("Use at() instead")

    def at(self, index: int) -> Statement:
        """This is similar to __getitem__ but due to the nature of the linked list,
        it is less efficient than __getitem__.

        Args:
            index (int): Index of the Statement.

        Returns:
            Statement: The Statement at the specified index.
        """
        if index >= len(self):
            raise IndexError("Index out of range")

        if index < 0:
            return self._at_reverse(-index - 1)

        return self._at_forward(index)

    def _at_forward(self, index: int) -> Statement:
        if self.node.first_stmt is None:
            raise IndexError("Index out of range")

        stmt = self.node.first_stmt
        for _ in range(index):
            if stmt is None:
                raise IndexError("Index out of range")
            stmt = stmt.next_stmt

        if stmt is None:
            raise IndexError("Index out of range")
        return stmt

    def _at_reverse(self, index: int) -> Statement:
        if self.node.last_stmt is None:
            raise IndexError("Index out of range")

        stmt = self.node.last_stmt
        for _ in range(index):
            if stmt is None:
                raise IndexError("Index out of range")
            stmt = stmt.prev_stmt

        if stmt is None:
            raise IndexError("Index out of range")
        return stmt

    def append(self, value: Statement) -> None:
        """Append a Statement to the reference Block.

        Args:
            value (Statement): A Statement to be appended.
        """
        from kirin.ir.nodes.stmt import Statement

        if not isinstance(value, Statement):
            raise ValueError(f"Expected Statement, got {type(value).__name__}")

        if self.node._stmt_len == 0:  # empty block
            value.attach(self.node)
            self.node._first_stmt = value
            self.node._last_stmt = value
            self.node._stmt_len += 1
        elif self.node._last_stmt:
            value.insert_after(self.node._last_stmt)
        else:
            raise ValueError("Invalid block, last_stmt is None")


@dataclass
class Block(IRNode["Region"]):
    """
    Block consist of a list of Statements and optionally input arguments.

    !!! note "Pretty Printing"
        This object is pretty printable via
        [`.print()`][kirin.print.printable.Printable.print] method.
    """

    _args: tuple[BlockArgument, ...]

    # NOTE: we need linked list since stmts are inserted frequently
    _first_stmt: Statement | None = field(repr=False)
    _last_stmt: Statement | None = field(repr=False)
    _stmt_len: int = field(default=0, repr=False)

    parent: Region | None = field(default=None, repr=False)
    """Parent Region of the Block."""

    def __init__(
        self,
        stmts: Sequence[Statement] = (),
        argtypes: Iterable[TypeAttribute] = (),
        *,
        source: SourceInfo | None = None,
    ):
        """
        Args:
            stmts (Sequence[Statement], optional): A list of statements. Defaults to ().
            argtypes (Iterable[TypeAttribute], optional): The type of the block arguments. Defaults to ().
        """
        super().__init__()
        self.source = source
        self._args = tuple(
            BlockArgument(self, i, argtype) for i, argtype in enumerate(argtypes)
        )

        self._first_stmt = None
        self._last_stmt = None
        self._first_branch = None
        self._last_branch = None
        self._stmt_len = 0
        self.stmts.extend(stmts)

    @property
    def parent_stmt(self) -> Statement | None:
        """parent statement of the Block."""
        if self.parent is None:
            return None
        return self.parent.parent_node

    @property
    def parent_node(self) -> Region | None:
        """Get parent Region of the Block."""
        return self.parent

    @parent_node.setter
    def parent_node(self, parent: Region | None) -> None:
        """Set the parent Region of the Block."""
        from kirin.ir.nodes.region import Region

        self.assert_parent(Region, parent)
        self.parent = parent

    @property
    def args(self) -> BlockArguments:
        """Get the  arguments of the Block.

        Returns:
            BlockArguments: The arguments view of the Block.
        """
        return BlockArguments(self, self._args)

    @property
    def first_stmt(self) -> Statement | None:
        """Get the first Statement of the Block.

        Returns:
            Statement | None: The first Statement of the Block.
        """
        return self._first_stmt

    @property
    def last_stmt(self) -> Statement | None:
        """Get the last Statement of the Block.

        Returns:
            Statement | None: The last Statement of the Block.

        """
        return self._last_stmt

    @property
    def stmts(self) -> BlockStmts:
        """Get the list of Statements of the Block.

        Returns:
            BlockStmts: The Statements of the Block.
        """
        return BlockStmts(self)

    def drop_all_references(self) -> None:
        """Remove all the dependency that reference/uses this Block."""
        self.parent = None
        for stmt in self.stmts:
            stmt.drop_all_references()

    def detach(self) -> None:
        """Detach this Block from the IR.

        Note:
            Detach only detach the Block from the IR graph. It does not remove uses that reference the Block.
        """
        if self.parent is None:
            return

        idx = self.parent[self]
        del self.parent._blocks[idx]
        del self.parent._block_idx[self]
        for block in self.parent._blocks[idx:]:
            self.parent._block_idx[block] -= 1
        self.parent = None

    def delete(self, safe: bool = True) -> None:
        """Delete the Block completely from the IR.

        Note:
            This method will detach + remove references of the block.

        Args:
            safe (bool, optional): If True, raise error if there is anything that still reference components in the block. Defaults to True.
        """
        self.detach()
        self.drop_all_references()
        for stmt in self.stmts:
            stmt.delete(safe=safe)

    def is_structurally_equal(
        self,
        other: Self,
        context: dict[IRNode | SSAValue, IRNode | SSAValue] | None = None,
    ) -> bool:
        """Check if the Block is structurally equal to another Block.

        Args:
            other (Self): The other Block to compare with.
            context (dict[IRNode  |  SSAValue, IRNode  |  SSAValue] | None, optional): A map of IRNode/SSAValue to hint that they are equivalent so the check will treat them as equivalent. Defaults to None.

        Returns:
            bool: True if the Block is structurally equal to the other Block.
        """
        if context is None:
            context = {}

        if len(self._args) != len(other._args) or len(self.stmts) != len(other.stmts):
            return False

        for arg, other_arg in zip(self._args, other._args):
            if arg.type != other_arg.type:
                return False
            context[arg] = other_arg

        context[self] = other
        if not all(
            stmt.is_structurally_equal(other_stmt, context)
            for stmt, other_stmt in zip(self.stmts, other.stmts)
        ):
            return False

        return True

    def __hash__(self) -> int:
        return id(self)

    def walk(
        self, *, reverse: bool = False, region_first: bool = False
    ) -> Iterator[Statement]:
        """Traversal the Statements in a Block.

        Args:
            reverse (bool, optional): If walk in the reversed manner. Defaults to False.
            region_first (bool, optional): If the walk should go through the Statement first or the Region of a Statement first. Defaults to False.

        Yields:
            Iterator[Statement]: An iterator that yield Statements in the Block in the specified order.
        """
        for stmt in reversed(self.stmts) if reverse else self.stmts:
            yield from stmt.walk(reverse=reverse, region_first=region_first)

    def print_impl(self, printer: Printer) -> None:
        printer.plain_print(printer.state.block_id[self])
        printer.print_seq(
            [printer.state.ssa_id[arg] for arg in self.args],
            delim=", ",
            prefix="(",
            suffix="):",
            emit=printer.plain_print,
        )

        if printer.analysis is not None:
            with printer.indent(increase=4, mark=False):
                for arg in self.args:
                    printer.print_newline()
                    with printer.rich(style="warning"):
                        printer.print_analysis(
                            arg, prefix=f"{printer.state.ssa_id[arg]} --> "
                        )

        with printer.indent(increase=2, mark=False):
            for stmt in self.stmts:
                printer.print_newline()
                printer.print_stmt(stmt)

    def verify(self) -> None:
        """Verify the correctness of the Block.

        Raises:
            ValidationError: If the Block is not correct.
        """
        from kirin.ir.nodes.stmt import Region

        if not isinstance(self.parent, Region):
            raise ValidationError(self, "Parent is not a region")

        for stmt in self.stmts:
            stmt.verify()

    def verify_type(self) -> None:
        """Verify the types of the Block.

        Raises:
            ValidationError: If the Block is not correct.
        """
        for stmt in self.stmts:
            stmt.verify_type()



from kirin import ir, types, lowering
from kirin.decl import info, statement

from ._dialect import dialect

T = types.TypeVar("T")


@statement
class BinOp(ir.Statement):
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    lhs: ir.SSAValue = info.argument(T, print=False)
    rhs: ir.SSAValue = info.argument(T, print=False)
    result: ir.ResultValue = info.result(T)


@statement(dialect=dialect)
class Add(BinOp):
    name = "add"


@statement(dialect=dialect)
class Sub(BinOp):
    name = "sub"


@statement(dialect=dialect)
class Mult(BinOp):
    name = "mult"


@statement(dialect=dialect)
class Div(BinOp):
    name = "div"


@statement(dialect=dialect)
class Mod(BinOp):
    name = "mod"


@statement(dialect=dialect)
class Pow(BinOp):
    name = "pow"


@statement(dialect=dialect)
class LShift(BinOp):
    name = "lshift"


@statement(dialect=dialect)
class RShift(BinOp):
    name = "rshift"


@statement(dialect=dialect)
class BitAnd(BinOp):
    name = "bitand"


@statement(dialect=dialect)
class BitOr(BinOp):
    name = "bitor"


@statement(dialect=dialect)
class BitXor(BinOp):
    name = "bitxor"


@statement(dialect=dialect)
class FloorDiv(BinOp):
    name = "floordiv"


@statement(dialect=dialect)
class MatMult(BinOp):
    name = "matmult"



from dataclasses import dataclass

from kirin.ir import IRNode
from kirin.rewrite.abc import RewriteRule, RewriteResult


@dataclass
class Fixpoint(RewriteRule):
    """Apply a rewrite rule until a fixpoint is reached.

    The rewrite rule is applied to the node until the rewrite rule does not do anything.

    ### Parameters
    - `map`: The rewrite rule to apply.
    - `max_iter`: The maximum number of iterations to apply the rewrite rule. Default is 32.
    """

    rule: RewriteRule
    max_iter: int = 32

    def rewrite(self, node: IRNode) -> RewriteResult:
        has_done_something = False
        for _ in range(self.max_iter):
            result = self.rule.rewrite(node)
            if result.terminated:
                return result

            if result.has_done_something:
                has_done_something = True
            else:
                return RewriteResult(has_done_something=has_done_something)

        return RewriteResult(exceeded_max_iter=True)



import ast
from dataclasses import dataclass


@dataclass
class SourceInfo:
    lineno: int
    col_offset: int
    end_lineno: int | None
    end_col_offset: int | None
    file: str | None = None
    lineno_begin: int = 0
    col_indent: int = 0

    @classmethod
    def from_ast(
        cls,
        node: ast.AST,
        file: str | None = None,
    ):
        end_lineno = getattr(node, "end_lineno", None)
        end_col_offset = getattr(node, "end_col_offset", None)
        return cls(
            getattr(node, "lineno", 0),
            getattr(node, "col_offset", 0),
            end_lineno if end_lineno is not None else None,
            end_col_offset if end_col_offset is not None else None,
            file,
        )

    def offset(self, lineno_begin: int = 0, col_indent: int = 0):
        """Offset the source info by the given offsets.

        Args:
            lineno_offset (int): The line number offset.
            col_offset (int): The column offset.
        """
        self.lineno_begin = lineno_begin
        self.col_indent = col_indent

    def __repr__(self) -> str:
        return (
            f'File "{self.file or "stdin"}", '
            f"line {self.lineno + self.lineno_begin},"
            f" col {self.col_offset + self.col_indent}"
        )



from types import MethodType as ClassMethodType, FunctionType

from kirin import ir, types
from kirin.decl import info, statement
from kirin.print.printer import Printer
from kirin.dialects.func.attrs import Signature, MethodType
from kirin.dialects.func.dialect import dialect

from .._pprint_helper import pprint_calllike


class FuncOpCallableInterface(ir.CallableStmtInterface["Function"]):

    @classmethod
    def get_callable_region(cls, stmt: "Function") -> ir.Region:
        return stmt.body


@statement(dialect=dialect)
class Function(ir.Statement):
    name = "func"
    traits = frozenset(
        {
            ir.IsolatedFromAbove(),
            ir.SymbolOpInterface(),
            ir.HasSignature(),
            FuncOpCallableInterface(),
            ir.SSACFGRegion(),
        }
    )
    sym_name: str = info.attribute()
    """The symbol name of the function."""
    signature: Signature = info.attribute()
    body: ir.Region = info.region(multi=True)

    def print_impl(self, printer: Printer) -> None:
        with printer.rich(style="keyword"):
            printer.print_name(self)
            printer.plain_print(" ")

        with printer.rich(style="symbol"):
            printer.plain_print(self.sym_name)

        printer.print_seq(self.signature.inputs, prefix="(", suffix=")", delim=", ")

        with printer.rich(style="comment"):
            printer.plain_print(" -> ")
            printer.print(self.signature.output)
            printer.plain_print(" ")

        printer.print(self.body)

        with printer.rich(style="comment"):
            printer.plain_print(f" // func.func {self.sym_name}")


@statement(dialect=dialect)
class Call(ir.Statement):
    name = "call"
    traits = frozenset({ir.MaybePure()})
    # not a fixed type here so just any
    callee: ir.SSAValue = info.argument()
    inputs: tuple[ir.SSAValue, ...] = info.argument()
    kwargs: tuple[str, ...] = info.attribute(default_factory=lambda: ())
    result: ir.ResultValue = info.result()
    purity: bool = info.attribute(default=False)

    def print_impl(self, printer: Printer) -> None:
        pprint_calllike(self, printer.state.ssa_id[self.callee], printer)

    def check_type(self) -> None:
        if not self.callee.type.is_subseteq(types.MethodType):
            if self.callee.type.is_subseteq(types.PyClass(FunctionType)):
                raise ir.TypeCheckError(
                    self,
                    f"callee must be a method type, got {self.callee.type}",
                    help="did you call a Python function directly? "
                    "consider decorating it with kernel decorator",
                )

            if self.callee.type.is_subseteq(types.PyClass(ClassMethodType)):
                raise ir.TypeCheckError(
                    self,
                    "callee must be a method type, got class method",
                    help="did you try to call a Python class method within a kernel? "
                    "consider rewriting it with a captured variable instead of calling it inside the kernel",
                )

            if self.callee.type is types.Any:
                return
            raise ir.TypeCheckError(
                self,
                f"callee must be a method type, got {self.callee.type}",
                help="did you forget to decorate the function with kernel decorator?",
            )


@statement(dialect=dialect)
class ConstantNone(ir.Statement):
    """A constant None value.

    This is mainly used to represent the None return value of a function
    to match Python semantics.
    """

    name = "const.none"
    traits = frozenset({ir.Pure(), ir.ConstantLike()})
    result: ir.ResultValue = info.result(types.NoneType)


@statement(dialect=dialect, init=False)
class Return(ir.Statement):
    name = "return"
    traits = frozenset({ir.IsTerminator(), ir.HasParent((Function,))})
    value: ir.SSAValue = info.argument()

    def __init__(self, value_or_stmt: ir.SSAValue | ir.Statement | None = None) -> None:
        if isinstance(value_or_stmt, ir.SSAValue):
            args = [value_or_stmt]
        elif isinstance(value_or_stmt, ir.Statement):
            if len(value_or_stmt._results) == 1:
                args = [value_or_stmt._results[0]]
            else:
                raise ValueError(
                    f"expected a single result, got {len(value_or_stmt._results)} results from {value_or_stmt.name}"
                )
        elif value_or_stmt is None:
            args = []
        else:
            raise ValueError(f"expected SSAValue or Statement, got {value_or_stmt}")

        super().__init__(args=args, args_slice={"value": 0})

    def print_impl(self, printer: Printer) -> None:
        with printer.rich(style="keyword"):
            printer.print_name(self)

        if self.args:
            printer.plain_print(" ")
            printer.print_seq(self.args, delim=", ")

    def check(self) -> None:
        assert self.args, "return statement must have at least one value"
        assert len(self.args) <= 1, (
            "return statement must have at most one value"
            ", wrap multiple values in a tuple"
        )


@statement(dialect=dialect)
class Lambda(ir.Statement):
    name = "lambda"
    traits = frozenset(
        {
            ir.Pure(),
            ir.HasSignature(),
            ir.SymbolOpInterface(),
            FuncOpCallableInterface(),
            ir.SSACFGRegion(),
        }
    )
    sym_name: str = info.attribute()
    signature: Signature = info.attribute()
    captured: tuple[ir.SSAValue, ...] = info.argument()
    body: ir.Region = info.region(multi=True)
    result: ir.ResultValue = info.result(MethodType)

    def check(self) -> None:
        assert self.body.blocks, "lambda body must not be empty"

    def print_impl(self, printer: Printer) -> None:
        with printer.rich(style="keyword"):
            printer.print_name(self)
        printer.plain_print(" ")

        with printer.rich(style="symbol"):
            printer.plain_print(self.sym_name)

        printer.print_seq(self.captured, prefix="(", suffix=")", delim=", ")

        with printer.rich(style="bright_black"):
            printer.plain_print(" -> ")
            printer.print(self.signature.output)

        printer.plain_print(" ")
        printer.print(self.body)

        with printer.rich(style="black"):
            printer.plain_print(f" // func.lambda {self.sym_name}")


@statement(dialect=dialect)
class GetField(ir.Statement):
    name = "getfield"
    traits = frozenset({ir.Pure()})
    obj: ir.SSAValue = info.argument(MethodType)
    field: int = info.attribute()
    # NOTE: mypy somehow doesn't understand default init=False
    result: ir.ResultValue = info.result(init=False)

    def print_impl(self, printer: Printer) -> None:
        printer.print_name(self)
        printer.plain_print(
            "(", printer.state.ssa_id[self.obj], ", ", str(self.field), ")"
        )
        with printer.rich(style="black"):
            printer.plain_print(" : ")
            printer.print(self.result.type)


@statement(dialect=dialect)
class Invoke(ir.Statement):
    name = "invoke"
    traits = frozenset({ir.MaybePure()})
    callee: ir.Method = info.attribute()
    inputs: tuple[ir.SSAValue, ...] = info.argument()
    kwargs: tuple[str, ...] = info.attribute()
    result: ir.ResultValue = info.result()
    purity: bool = info.attribute(default=False)

    def print_impl(self, printer: Printer) -> None:
        pprint_calllike(self, self.callee.sym_name, printer)

    def check(self) -> None:
        if self.kwargs:
            for name in self.kwargs:
                assert (
                    name in self.callee.arg_names
                ), f"method {self.callee.sym_name} does not have argument {name}"
        elif len(self.callee.arg_names) - 1 != len(self.args):
            raise ValueError(
                f"expected {len(self.callee.arg_names)} arguments, got {len(self.args)}"
            )



import inspect
import textwrap

from beartype.door import is_subhint

from kirin import ir

from .base import BaseModifier
from .info import Field, ResultField, ArgumentField


class Verify(BaseModifier):
    RESERVED_NAMES = {
        "name",
        "traits",
        "dialect",
        "args",
        "parent",
        "parent_node",
        "parent_region",
        "parent_block",
        "next_stmt",
        "prev_stmt",
        "results",
        "regions",
        "successors",
        "attributes",
    }

    def verify_mro(self):
        if not issubclass(self.cls, ir.Statement):
            raise TypeError(
                f"expected {self.cls.__name__} to be a subclass of ir.Statement"
            )

    def verify_fields(self):
        # Do we have any Field members that don't also have annotations?
        cls_annotations = inspect.get_annotations(self.cls)
        for name, value in self.cls.__dict__.items():
            if isinstance(value, Field) and name not in cls_annotations:
                raise ValueError(f"{name!r} is a field but has no type annotation")

        for f in self.fields:
            if f.name in self.RESERVED_NAMES:
                raise ValueError(f"{f.name!r} is a reserved name")

            if isinstance(f, ArgumentField):
                if not (
                    is_subhint(f.annotation, ir.SSAValue)
                    or is_subhint(f.annotation, tuple[ir.SSAValue, ...])
                ):
                    raise ValueError(
                        f"{f.name!r} is an argument field but has an invalid type annotation"
                    )

                if is_subhint(f.annotation, ir.ResultValue):
                    raise ValueError(
                        textwrap.dedent(
                            f"{f.name!r} is an argument field but has an invalid type annotation {f.annotation}"
                            " (did you mean to use `info.result` instead?)"
                        )
                    )

            if isinstance(f, ResultField) and not is_subhint(
                f.annotation, ir.ResultValue
            ):
                raise ValueError(
                    f"{f.name!r} is a result field but has an invalid type annotation"
                )



from kirin import ir
from kirin.prelude import python_basic
from kirin.dialects import func, ilist, lowering


@ir.dialect_group(python_basic.union([func, ilist, lowering.func]))
def basic_desugar(self):
    ilist_desugar = ilist.IListDesugar(self)

    def run_pass(
        mt: ir.Method,
    ) -> None:
        ilist_desugar(mt)

    return run_pass


def test_ilist2list_rewrite():

    x = [1, 2, 3, 4]

    @basic_desugar
    def ilist2_list():
        return x

    ilist2_list.print()

    x = ilist2_list()

    assert isinstance(x, ilist.IList)



from kirin import ir, types
from kirin.rewrite import Walk, Chain, Fixpoint
from kirin.passes.abc import Pass
from kirin.rewrite.abc import RewriteResult
from kirin.dialects.ilist.rewrite import List2IList, ConstList2IList


class IListDesugar(Pass):
    """This pass desugars the Python list dialect
    to the immutable list dialect by rewriting all
    constant `list` type into `IList` type.
    """

    def unsafe_run(self, mt: ir.Method) -> RewriteResult:
        for arg in mt.args:
            _check_list(arg.type, arg.type)
        return Fixpoint(Walk(Chain(ConstList2IList(), List2IList()))).rewrite(mt.code)


def _check_list(total: types.TypeAttribute, type_: types.TypeAttribute):
    if isinstance(type_, types.Generic):
        _check_list(total, type_.body)
        for var in type_.vars:
            _check_list(total, var)
        if type_.vararg:
            _check_list(total, type_.vararg.typ)
    elif isinstance(type_, types.PyClass):
        if issubclass(type_.typ, list):
            raise TypeError(
                f"Invalid type {total} for this kernel, use IList instead of {type_}."
            )
    return



from typing import Any

from kirin import ir, types
from kirin.decl import info, fields, statement
from kirin.decl.info import (
    BlockField,
    RegionField,
    ResultField,
    ArgumentField,
    AttributeField,
)

T = types.TypeVar("T")


@statement
class DummyStatement(ir.Statement):
    name = "constant"
    traits = frozenset({ir.Pure(), ir.ConstantLike()})

    # args
    noinfo: ir.SSAValue
    vararg_noinfo: tuple[ir.SSAValue, ...]
    xxx: ir.SSAValue = info.argument(T)
    xxx_Any: ir.SSAValue = info.argument()
    # xxx_vararg: list[ir.SSAValue] = argument()

    # results
    xxx_result: ir.ResultValue = info.result(T)

    # attributes
    xxx_attribute: Any = info.attribute(T)
    xxx_dict: dict[str, int] = info.attribute()

    # regions
    xxx_region_noinfo: ir.Region = info.region()
    xxx_region: ir.Region = info.region(default_factory=ir.Region)
    xxx_region_multi: ir.Region = info.region(default_factory=ir.Region, multi=True)

    # blocks
    block_noinfo: ir.Block = info.block()
    block_default: ir.Block = info.block(default_factory=ir.Block)


ff = fields(DummyStatement)


def test_scan_fields():
    noinfo = ff.args["noinfo"]
    assert isinstance(noinfo, ArgumentField)
    assert noinfo.type == types.Any
    assert noinfo.group is False

    vararg_noinfo = ff.args["vararg_noinfo"]
    assert isinstance(vararg_noinfo, ArgumentField)
    assert vararg_noinfo.type == types.Any
    assert vararg_noinfo.group is True

    xxx = ff.args["xxx"]
    assert isinstance(xxx, ArgumentField)
    assert xxx.type == T
    assert xxx.group is False

    xxx_Any = ff.args["xxx_Any"]
    assert isinstance(xxx_Any, ArgumentField)
    assert xxx_Any.type == types.Any
    assert xxx_Any.group is False

    xxx_result = ff.results["xxx_result"]
    assert isinstance(xxx_result, ResultField)
    assert xxx_result.type == T

    xxx_attribute = ff.attributes["xxx_attribute"]
    assert isinstance(xxx_attribute, AttributeField)
    assert xxx_attribute.type == T

    xxx_dict = ff.attributes["xxx_dict"]
    assert isinstance(xxx_dict, AttributeField)
    assert xxx_dict.type == types.Dict[types.String, types.Int]

    xxx_region_noinfo = ff.regions["xxx_region_noinfo"]
    assert isinstance(xxx_region_noinfo, RegionField)
    assert xxx_region_noinfo.multi is False

    xxx_region = ff.regions["xxx_region"]
    assert isinstance(xxx_region, RegionField)
    assert xxx_region.multi is False

    xxx_region_multi = ff.regions["xxx_region_multi"]
    assert isinstance(xxx_region_multi, RegionField)
    assert xxx_region_multi.multi is True

    block_noinfo = ff.blocks["block_noinfo"]
    assert isinstance(block_noinfo, BlockField)

    block = ff.blocks["block_default"]
    assert isinstance(block, BlockField)



from kirin.decl.base import BaseModifier

from ._set_new_attribute import set_new_attribute


class EmitTraits(BaseModifier):

    def emit_traits(self):
        # if no parent defines traits, set it to empty set
        for base in self.cls.__mro__[-1:0:-1]:
            if hasattr(base, "traits"):
                return
        set_new_attribute(self.cls, "traits", frozenset({}))



from dataclasses import dataclass

from kirin.ir import Pure, Block, Statement
from kirin.rewrite.abc import RewriteRule, RewriteResult


@dataclass
class CommonSubexpressionElimination(RewriteRule):

    def rewrite_Block(self, node: Block) -> RewriteResult:
        seen: dict[int, Statement] = {}

        for stmt in node.stmts:
            if not stmt.has_trait(Pure):
                continue

            if stmt.regions:
                continue

            hash_value = hash(
                (type(stmt),)
                + tuple(stmt.args)
                + tuple(stmt.attributes.values())
                + tuple(stmt.successors)
                + tuple(stmt.regions)
            )
            if hash_value in seen:
                old_stmt = seen[hash_value]
                for result, old_result in zip(stmt._results, old_stmt.results):
                    result.replace_by(old_result)
                stmt.delete()
                return RewriteResult(has_done_something=True)
            else:
                seen[hash_value] = stmt
        return RewriteResult()

    def rewrite_Statement(self, node: Statement) -> RewriteResult:
        if not node.regions:
            return RewriteResult()

        has_done_something = False
        for region in node.regions:
            for block in region.blocks:
                result = self.rewrite_Block(block)
                if result.has_done_something:
                    has_done_something = True

        return RewriteResult(has_done_something=has_done_something)



from typing import TypeVar
from dataclasses import dataclass

from kirin.print import Printer

from .data import Data
from .types import PyClass, TypeAttribute

T = TypeVar("T")


@dataclass
class PyAttr(Data[T]):
    """Python attribute for compile-time values.
    This is a generic attribute that holds a Python value.

    The constructor takes a Python value and an optional type attribute.
    If the type attribute is not provided, the type of the value is inferred
    as `PyClass(type(value))`.

    !!! note "Pretty Printing"
        This object is pretty printable via
        [`.print()`][kirin.print.printable.Printable.print] method.
    """

    name = "PyAttr"
    data: T

    def __init__(self, data: T, pytype: TypeAttribute | None = None):
        self.data = data

        if pytype is None:
            self.type = PyClass(type(data))
        else:
            self.type = pytype

    def __hash__(self):
        # Fix hash(-1) == hash(-2) collision
        # assume maximum is 8 bytes == 64 bits
        if isinstance(self.data, int):
            return hash(self.data.to_bytes(signed=True, byteorder="big", length=8))
        elif isinstance(self.data, float):
            return hash(self.data.hex())
        return hash(self.data) + hash(self.type)

    def print_impl(self, printer: Printer) -> None:
        printer.plain_print(repr(self.data))
        with printer.rich(style="comment"):
            printer.plain_print(" : ")
            printer.print(self.type)

    def unwrap(self) -> T:
        return self.data



from kirin import lowering
from kirin.prelude import basic_no_opt
from kirin.dialects import func
from kirin.analysis.cfg import CFG

lower = lowering.Python(basic_no_opt)


def deadblock(x):
    if x:
        return x + 1
    else:
        return x + 2
    return x + 3


def test_reachable():
    code = lower.python_function(deadblock, compactify=False)
    assert isinstance(code, func.Function)
    cfg = CFG(code.body)
    assert code.body.blocks[-1] not in cfg.successors


def foo(x: int):  # type: ignore
    def goo(y: int):
        return x + y

    return goo


def test_foo_cfg():
    code = lower.python_function(foo, compactify=False)
    assert isinstance(code, func.Function)
    cfg = CFG(code.body)
    assert code.body.blocks[0] in cfg.successors
    assert code.body.blocks[1] not in cfg.successors



from typing import TYPE_CHECKING, Any, Generic, TypeVar, Iterable, Optional, Protocol

if TYPE_CHECKING:
    from kirin import ir
    from kirin.print import Printer

Node = TypeVar("Node")


class Graph(Protocol, Generic[Node]):
    """The graph interface.

    This interface defines the methods that a graph object must implement.
    The graph interface is mainly for compatibility reasons so that one can
    use multiple graph implementations interchangeably.
    """

    def get_neighbors(self, node: Node) -> Iterable[Node]:
        """Get the neighbors of a node."""
        ...

    def get_nodes(self) -> Iterable[Node]:
        """Get all the nodes in the graph."""
        ...

    def get_edges(self) -> Iterable[tuple[Node, Node]]:
        """Get all the edges in the graph."""
        ...

    def print(
        self,
        printer: Optional["Printer"] = None,
        analysis: dict["ir.SSAValue", Any] | None = None,
    ) -> None: ...



from typing import TypeVar
from dataclasses import field, dataclass

from kirin import ir, types, interp
from kirin.worklist import WorkList

ValueType = TypeVar("ValueType")

@dataclass
class EmitFrame(interp.Frame[ValueType]):
    worklist: WorkList[interp.Successor] = field(default_factory=WorkList)
    block_ref: dict[ir.Block, ValueType] = field(default_factory=dict)

FrameType = TypeVar("FrameType", bound=EmitFrame)

class EmitABC(interp.BaseInterpreter[FrameType, ValueType]):
    def run_callable_region(
        self,
        frame: FrameType,
        code: ir.Statement,
        region: ir.Region,
        args: tuple[ValueType, ...],
    ) -> ValueType: ...
    def run_ssacfg_region(
        self, frame: FrameType, region: ir.Region, args: tuple[ValueType, ...]
    ) -> tuple[ValueType, ...]: ...
    def emit_attribute(self, attr: ir.Attribute) -> ValueType: ...
    def emit_type_Any(self, attr: types.AnyType) -> ValueType: ...
    def emit_type_Bottom(self, attr: types.BottomType) -> ValueType: ...
    def emit_type_Literal(self, attr: types.Literal) -> ValueType: ...
    def emit_type_Union(self, attr: types.Union) -> ValueType: ...
    def emit_type_TypeVar(self, attr: types.TypeVar) -> ValueType: ...
    def emit_type_Vararg(self, attr: types.Vararg) -> ValueType: ...
    def emit_type_Generic(self, attr: types.Generic) -> ValueType: ...
    def emit_type_PyClass(self, attr: types.PyClass) -> ValueType: ...
    def emit_type_PyAttr(self, attr: ir.PyAttr) -> ValueType: ...
    def emit_attribute_fallback(self, attr: ir.Attribute) -> ValueType: ...
    def emit_stmt_begin(self, frame: FrameType, stmt: ir.Statement) -> None: ...
    def emit_stmt_end(self, frame: FrameType, stmt: ir.Statement) -> None: ...
    def emit_block_begin(self, frame: FrameType, block: ir.Block) -> None: ...
    def emit_block_end(self, frame: FrameType, block: ir.Block) -> None: ...
    def emit_block(self, frame: FrameType, block: ir.Block) -> ValueType: ...



import pytest

from kirin import ir
from kirin.prelude import basic


def test_main():
    y = 1

    @basic
    def foo(x):
        return x + y

    with pytest.raises(ir.CompilerError):

        @basic
        def foo(x):  # noqa: F811
            return x + y



from kirin.prelude import basic_no_opt
from kirin.passes.inline import InlinePass


@basic_no_opt
def inline_func(x: int):
    return x - 1


def test_inline_pass():

    @basic_no_opt
    def main_inline_pass(x: int):
        y = inline_func(x)
        return y + 1

    inline = InlinePass(main_inline_pass.dialects)
    a = main_inline_pass(1)
    main_inline_pass.code.print()
    inline(main_inline_pass)
    main_inline_pass.code.print()
    b = main_inline_pass(1)
    assert a == b
    assert len(main_inline_pass.callable_region.blocks[0].stmts) == 5


def test_inline_pass_custom_heru():

    @basic_no_opt
    def main_inline_pass2(x: int):
        y = inline_func(x)
        return y + 1

    inline = InlinePass(main_inline_pass2.dialects, herustic=lambda x: False)
    a = main_inline_pass2(1)
    main_inline_pass2.code.print()
    inline(main_inline_pass2)
    main_inline_pass2.code.print()
    b = main_inline_pass2(1)
    assert a == b

    assert len(main_inline_pass2.callable_region.blocks[0].stmts) == 4



"""IR module for kirin.

This module contains the data structure definition
of the intermediate representation (IR) for kirin.
"""

from kirin.ir import attrs as attrs, traits as traits
from kirin.ir.ssa import (
    SSAValue as SSAValue,
    TestValue as TestValue,
    ResultValue as ResultValue,
    BlockArgument as BlockArgument,
    DeletedSSAValue as DeletedSSAValue,
)
from kirin.ir.use import Use as Use
from kirin.ir.group import DialectGroup as DialectGroup, dialect_group as dialect_group
from kirin.ir.nodes import (
    Block as Block,
    IRNode as IRNode,
    Region as Region,
    Statement as Statement,
)
from kirin.ir.method import Method as Method
from kirin.ir.traits import (
    Pure as Pure,
    Trait as Trait,
    AttrTrait as AttrTrait,
    HasParent as HasParent,
    MaybePure as MaybePure,
    StmtTrait as StmtTrait,
    RegionTrait as RegionTrait,
    SymbolTable as SymbolTable,
    ConstantLike as ConstantLike,
    HasSignature as HasSignature,
    IsTerminator as IsTerminator,
    NoTerminator as NoTerminator,
    SSACFGRegion as SSACFGRegion,
    IsolatedFromAbove as IsolatedFromAbove,
    SymbolOpInterface as SymbolOpInterface,
    CallableStmtInterface as CallableStmtInterface,
)
from kirin.ir.dialect import Dialect as Dialect
from kirin.ir.attrs.py import PyAttr as PyAttr
from kirin.ir.attrs.abc import Attribute as Attribute, AttributeMeta as AttributeMeta
from kirin.ir.exception import (
    CompilerError as CompilerError,
    TypeCheckError as TypeCheckError,
    ValidationError as ValidationError,
)
from kirin.ir.attrs.data import Data as Data



from kirin import ir, types, lowering
from kirin.decl import info, statement

from ._dialect import dialect

T = types.TypeVar("T")


@statement(dialect=dialect)
class New(ir.Statement):
    name = "list"
    traits = frozenset({lowering.FromPythonCall()})
    values: tuple[ir.SSAValue, ...] = info.argument(T)
    result: ir.ResultValue = info.result(types.List[T])


@statement(dialect=dialect)
class Append(ir.Statement):
    name = "append"
    traits = frozenset({lowering.FromPythonCall()})
    list_: ir.SSAValue = info.argument(types.List[T])
    value: ir.SSAValue = info.argument(T)
